# %%
# pyright: reportWildcardImportFromLibrary=false
import dataclasses
import math
import jax
import jax.numpy as jnp
import pytest
from minigenjax import *


@Gen
def model1(b):
    y = Normal(b, 0.1) @ "x"
    return y


@Gen
def model2(b):
    return Uniform(b, b + 2.0) @ "x"


@Gen
def model3(x):
    a = model1(x) @ "a"
    b = model2(x / 2.0) @ "b"
    return a, b


@Gen
def cond_model(b):
    flip = Flip(0.5) @ "flip"
    y = Cond(model1(b), model2(b / 2.0))(flip) @ "s"
    return y


@Gen
def inlier_model(y, sigma_inlier):
    return Normal(y, sigma_inlier) @ "value"


@Gen
def outlier_model(y):
    return Uniform(y - 1.0, y + 1.0) @ "value"


@Gen
def curve_model(f, x, p_outlier, sigma_inlier):
    outlier = Flip(p_outlier) @ "outlier"
    y = f(x)
    fork = Cond(outlier_model(y), inlier_model(y, sigma_inlier))
    return fork(outlier) @ "y"


@Gen
def coefficient():
    return Normal(0.0, 1.0) @ "c"


@jax.tree_util.register_dataclass
@dataclasses.dataclass
class Poly:
    coefficients: jax.Array

    def __call__(self, x):
        if not self.coefficients.shape:
            return 0.0
        powers = jnp.pow(
            jnp.array(x)[jnp.newaxis], jnp.arange(self.coefficients.shape[0])
        )
        return self.coefficients.T @ powers


key0 = jax.random.key(0)


def test_pytree():
    poly = coefficient().repeat(3).map(Poly)
    tr = poly.simulate(key0)
    p = tr["retval"]
    assert jnp.allclose(p.coefficients, jnp.array([1.1188384, 0.5781488, 0.8535516]))

    @Gen
    def noisy_eval(f, x):
        return f(x) + Normal(0.0, 0.01) @ "noise"

    tr = jax.jit(noisy_eval(p, 0.0).simulate)(jax.random.key(1))
    assert tr["retval"] == 1.1069956
    assert tr["retval"] == p(0.0) + tr["subtraces"]["noise"]["retval"]

    tr = jax.jit(noisy_eval.vmap(in_axes=(None, 0))(p, jnp.arange(-2.0, 2.0)).simulate)(
        jax.random.key(2)
    )
    assert jnp.allclose(
        tr["retval"], jnp.array([3.37815, 1.3831037, 1.1251557, 2.533188])
    )


# %%
def test_normal_model():
    tr = model1(10.0).simulate(key0)
    expected = {
        "retval": 9.979416,
        "subtraces": {
            "x": {
                "retval": jnp.array(9.979416),
                "score": jnp.array(1.3624613),
            }
        },
    }
    assert tr == expected


def test_uniform_model():
    tr = model2(20.0).simulate(key0)
    assert tr == {
        "retval": 20.836914,
        "subtraces": {
            "x": {
                "retval": 20.836914,
                "score": -0.6931472,
            }
        },
    }


def test_multiple_results():
    tr = model3(50.0).simulate(key0)
    assert tr["retval"] == (49.874847, 26.114412)


def test_logit_vs_probs():
    def sigmoid(g):
        return math.exp(g) / (1.0 + math.exp(g))

    @Gen
    def model():
        g = Bernoulli(logits=-0.3) @ "l"
        p = Bernoulli(probs=0.3) @ "p"
        return g, p

    tr = model().simulate(key0)
    print(jax.make_jaxpr(model().simulate)(key0))
    assert tr["subtraces"]["l"]["retval"] == 1.0
    assert (
        tr["subtraces"]["l"]["score"]
        == -0.8543553
        == pytest.approx(math.log(sigmoid(-0.3)))
    )
    assert tr["subtraces"]["p"]["retval"] == 0.0
    assert (
        tr["subtraces"]["p"]["score"] == -0.35667497 == pytest.approx(math.log(1 - 0.3))
    )


def test_model_vmap():
    tr = jax.vmap(model3(50.0).map(sum).simulate)(jax.random.split(key0, 5))
    assert jnp.allclose(
        tr["retval"], jnp.array([76.4031, 76.777, 75.255844, 76.623726, 76.145515])
    )
    assert jnp.allclose(
        tr["subtraces"]["a"]["subtraces"]["x"]["retval"],
        jnp.array([49.966385, 50.01686, 50.021904, 50.22247, 49.879295]),
    )
    assert jnp.allclose(
        tr["subtraces"]["b"]["subtraces"]["x"]["retval"],
        jnp.array([26.436712, 26.760138, 25.233942, 26.401255, 26.26622]),
    )
    assert jnp.allclose(
        tr["subtraces"]["a"]["subtraces"]["x"]["score"],
        jnp.array([1.3271478, 1.369432, 1.3596607, -1.0910004, 0.65514755]),
    )
    assert jnp.allclose(
        tr["subtraces"]["b"]["subtraces"]["x"]["score"],
        jnp.array([-0.6931472, -0.6931472, -0.6931472, -0.6931472, -0.6931472]),
    )


def test_distribution_as_sampler():
    def vmap(n):
        return lambda f: jax.vmap(f)(jax.random.split(key0, n))

    assert jnp.allclose(
        vmap(10)(Normal(0.0, 0.01)),
        jnp.array(
            [
                -0.00449334,
                -0.00115321,
                -0.005181,
                0.00307154,
                -0.02684483,
                -0.01266131,
                0.00193166,
                -0.01589755,
                -0.00339408,
                0.01838289,
            ]
        ),
    )
    assert jnp.allclose(
        vmap(10)(Uniform(5.0, 6.0)),
        jnp.array(
            [
                5.3265953,
                5.454095,
                5.302194,
                5.620637,
                5.003632,
                5.102733,
                5.576586,
                5.055945,
                5.3671513,
                5.96699,
            ]
        ),
    )
    assert jnp.allclose(
        vmap(10)(Flip(0.5)),
        jnp.array([1, 1, 1, 0, 1, 1, 0, 1, 1, 0]),
    )
    assert jnp.allclose(
        vmap(10)(Categorical(logits=jnp.array([1.1, -1.0, 0.9]))),
        jnp.array([0, 0, 2, 2, 0, 1, 0, 0, 0, 0]),
    )
    assert jnp.allclose(
        vmap(10)(MvNormalDiag(jnp.array([1.0, 10.0, 100.0]), 0.1 * jnp.ones(3))),
        jnp.array(
            [
                [1.0678201, 9.900013, 100.02386],
                [1.0617225, 10.046055, 99.96553],
                [0.97022116, 10.116703, 100.034744],
                [0.8233146, 10.12513, 100.1254],
                [0.9622561, 9.9325695, 99.90703],
                [0.827492, 10.113411, 99.91007],
                [1.1142846, 10.202795, 99.900986],
                [0.9468851, 9.946305, 99.93078],
                [0.9302361, 9.9734125, 99.82939],
                [1.0136702, 9.983562, 99.88976],
            ]
        ),
    )


def test_cond_model():
    b = 100.0
    tr = model1(b).simulate(key0)
    assert jnp.allclose(tr["retval"], jnp.array(99.979416))
    tr = model2(b / 2).simulate(key0)
    assert jnp.allclose(tr["retval"], jnp.array(50.836914))
    c = Cond(model1(b), model2(b / 2.0))
    tr = c(0).simulate(key0)
    assert jnp.allclose(tr["retval"], jnp.array(50.836914))
    tr = c(1).simulate(key0)
    assert jnp.allclose(tr["retval"], jnp.array(99.979416))
    tr = jax.vmap(lambda i, k: c(i).simulate(k))(
        jnp.mod(jnp.arange(10.0), 2), jax.random.split(key0, 10)
    )
    assert jnp.allclose(
        tr["retval"],
        jnp.array(
            [
                50.65319,
                99.988464,
                50.60439,
                100.030716,
                50.007263,
                99.87339,
                51.15317,
                99.84103,
                50.734303,
                100.18383,
            ]
        ),
    )

    @Gen
    def cond_model(b):
        flip = Flip(0.5) @ "flip"
        return c(flip) @ "c"

    tr = cond_model(b).simulate(key0)
    assert jnp.allclose(
        tr["subtraces"]["c"]["subtraces"]["x"]["retval"], jnp.array(100.01439)
    )
    assert jnp.allclose(tr["subtraces"]["flip"]["retval"], jnp.array(1))


def test_vmap_over_cond():
    tr = jax.vmap(cond_model(100.0).simulate)(jax.random.split(key0, 5))

    assert jnp.allclose(
        tr["retval"], jnp.array([100.0578, 51.76014, 50.23394, 51.401253, 100.03401])
    )
    assert jnp.allclose(tr["subtraces"]["flip"]["retval"], jnp.array([1, 0, 0, 0, 1]))
    assert jnp.allclose(
        tr["subtraces"]["s"]["retval"],
        jnp.array([100.0578, 51.76014, 50.23394, 51.401253, 100.03401]),
    )


def test_ordinary_cond():
    @Gen
    def f():
        n = Normal(0.0, 1.0) @ "n"
        return jax.lax.cond(n > 0, lambda: n, lambda: 10 * n)

    tr = f().simulate(key0)
    assert tr["retval"] == -12.5153885
    assert tr["subtraces"]["n"]["retval"] == -1.2515389


def test_intervening_functions():
    @Gen
    def h():
        return Normal(0.0, 1.0) @ "n"

    def g():
        return h()

    @Gen
    def f():
        return g() @ "g"

    tr = f().simulate(key0)
    assert tr["retval"] == -0.20584226
    assert tr["subtraces"]["g"]["retval"] == tr["retval"]


def test_scan_model():
    @Gen
    def update(state, delta):
        drift = Normal(delta, 0.01) @ "drift"
        new_position = state + drift
        return new_position, new_position

    @Gen
    def scan_update():
        return Scan(update)(10.0, jnp.arange(0.1, 0.6, 0.1)) @ "S"

    tr = scan_update().simulate(key0)
    assert jnp.allclose(tr["retval"][0], 11.482168)
    assert jnp.allclose(
        tr["retval"][1],
        jnp.array([10.087484, 10.281618, 10.586483, 10.988654, 11.482168]),
    )
    assert jnp.allclose(
        tr["subtraces"]["S"]["subtraces"]["drift"]["retval"],
        jnp.array([0.08748461, 0.19413349, 0.30486485, 0.4021714, 0.49351367]),
    )
    assert jnp.allclose(
        tr["subtraces"]["S"]["subtraces"]["drift"]["score"],
        jnp.array([2.903057, 3.514152, 3.5678985, 3.6626565, 3.4758694]),
    )


def test_plain_scan():
    @Gen
    def model(x):
        init = Normal(x, 0.01) @ "init"
        return jax.lax.scan(lambda a, b: (a + b, a + b), init, jnp.arange(5.0))

    tr = model(10.0).simulate(key0)
    assert tr["retval"][0] == 19.987484
    assert jnp.allclose(
        tr["retval"][1],
        jnp.array([9.987485, 10.987485, 12.987485, 15.987485, 19.987484]),
    )
    assert tr["subtraces"]["init"]["retval"] == 9.987485


class TestCurve:
    def test_curve_model(self):
        f = Poly(jnp.array([1.0, -1.0, 2.0]))  # x**2.0 - x + 1.0

        assert f(0.0) == 1.0

        tr = curve_model(f, 0.0, 0.0, 0.0).simulate(key0)
        assert tr["retval"] == 1.0
        assert tr["subtraces"]["outlier"]["retval"] == 0
        assert tr["subtraces"]["y"]["subtraces"]["value"]["retval"] == 1.0

        tr = curve_model.vmap(in_axes=(None, 0, None, None))(
            f, jnp.arange(-3.0, 3.0), 0.01, 0.01
        ).simulate(key0)
        assert jnp.allclose(
            tr["subtraces"]["outlier"]["retval"],
            jnp.array([0, 0, 0, 0, 0, 0]),
        )
        assert jnp.allclose(
            tr["retval"],
            jnp.array([22.00005, 10.993716, 3.9880881, 0.99697554, 1.99467, 7.027452]),
        )
        tr = curve_model.vmap(in_axes=(None, None, 0, None))(
            f, 0.0, jnp.array([0.001, 0.01, 0.9]), 0.3
        ).simulate(key0)
        assert jnp.allclose(tr["subtraces"]["outlier"]["retval"], jnp.array([0, 0, 1]))
        assert jnp.allclose(tr["retval"], jnp.array([0.9980389, 0.91635126, 1.0424924]))

    def test_curve_generation(self):
        quadratic = coefficient().repeat(3).map(Poly)
        points = jnp.arange(-3, 4) / 10.0

        # curve_model(f, x, p_outlier, sigma_inlier)
        @Gen
        def model(xs):
            poly = quadratic @ "p"
            p_outlier = Uniform(0.0, 1.0) @ "p_outlier"
            sigma_inlier = Uniform(0.0, 0.3) @ "sigma_inlier"
            return (
                curve_model.vmap(in_axes=(None, 0, None, None))(
                    poly, xs, p_outlier, sigma_inlier
                )
                @ "y"
            )

        # print(jax.make_jaxpr(model(points).simulate)(key0))
        jit_model = jax.jit(model(points).simulate)

        tr = jit_model(key0)
        assert jnp.allclose(
            tr["subtraces"]["p"]["subtraces"]["c"]["retval"],
            jnp.array([0.785558, 2.3734226, 0.07902155]),
        )
        assert jnp.allclose(
            tr["retval"],
            jnp.array(
                [
                    -0.4003628,
                    0.13468444,
                    0.43138668,
                    1.0624609,
                    0.82224107,
                    1.1899176,
                    1.8036526,
                ]
            ),
        )
        assert jnp.allclose(
            tr["subtraces"]["y"]["subtraces"]["outlier"]["retval"],
            jnp.array([1.0, 1.0, 0.0, 0.0, 1.0, 0.0, 1.0]),
        )
        assert jnp.allclose(
            tr["retval"],
            jnp.array(
                [
                    -0.4003628,
                    0.13468444,
                    0.43138668,
                    1.0624609,
                    0.82224107,
                    1.1899176,
                    1.8036526,
                ]
            ),
        )

        tr = jax.vmap(jit_model)(jax.random.split(key0, 10))
        assert tr["subtraces"]["p"]["subtraces"]["c"]["retval"].shape == (10, 3)
        assert tr["retval"].shape == (10, 7)


def test_map():
    @Gen
    def noisy(x):
        return Normal(x, 0.01) @ "x"

    def plus5(x):
        return x + 5.0

    noisy_plus5 = noisy(10).map(plus5)
    tr = jax.vmap(noisy_plus5.simulate)(jax.random.split(key0, 3))
    assert jnp.allclose(tr["retval"], jnp.array([15.0111885, 15.005781, 15.008535]))


def test_simple_repeat():
    @Gen
    def coefficient():
        return Normal(0.0, 1.0) @ "c"

    def Poly(coefficient_gf, n):
        @Gen
        def poly():
            return coefficient_gf().repeat(n) @ "cs"

        return poly

    poly4 = Poly(coefficient, 4)
    tr = poly4().simulate(key0)
    assert jnp.allclose(
        tr["retval"], jnp.array([0.30984825, -1.3642794, 2.2861156, 0.6714109])
    )


def test_repeat_in_model():
    @Gen
    def x(y):
        return Normal(2.0 * y, 1.0) @ "x"

    @Gen
    def xs():
        return x(10.0).repeat(4) @ "xs"

    tr = xs().simulate(key0)
    assert jnp.allclose(
        tr["retval"], jnp.array([20.309849, 18.635721, 22.286116, 20.671412])
    )


def test_repeat_of_repeat():
    @Gen
    def y(x):
        return Normal(2.0 * x + 1, 0.1) @ "y"

    tr = y(5.0).repeat(4).repeat(3).simulate(key0)
    assert jnp.allclose(
        tr["retval"],
        jnp.array(
            [
                [10.956121, 11.274088, 10.94439, 10.894012],
                [11.076275, 11.013761, 11.081145, 10.860387],
                [11.015622, 11.002483, 10.953617, 10.820805],
            ]
        ),
    )


def test_map_in_model():
    @Gen
    def x(y):
        return Normal(y, 0.1) @ "x"

    @Gen
    def mx():
        return x(7.0).map(lambda t: t + 13.0) @ "mx"

    tr = jax.vmap(mx().simulate)(jax.random.split(key0, 5))
    assert jnp.allclose(
        tr["retval"], jnp.array([19.907951, 19.926113, 20.161331, 19.842987, 19.94448])
    )


def test_map_of_repeat():
    @Gen
    def coefficient():
        return Normal(0.0, 1.0) @ "c"

    pg = coefficient().repeat(3).map(Poly)

    tr = pg.simulate(key0)
    assert jnp.allclose(
        tr["retval"].coefficients, jnp.array([1.1188384, 0.5781488, 0.8535516])
    )
    assert jnp.allclose(tr["retval"](1.0), jnp.array(2.5505388))
    assert jnp.allclose(tr["retval"](2.0), jnp.array(5.6893425))

    kg = coefficient().repeat(3).map(jnp.sum)
    tr = kg.simulate(key0)
    assert jnp.allclose(tr["retval"], jnp.array(2.5505388))


def test_repeat_of_map():
    @Gen
    def y(x):
        return Normal(x, 0.1) @ "y"

    mr = y(7.0).map(lambda x: x + 13.0).repeat(5)

    tr = mr.simulate(key0)
    assert jnp.allclose(
        tr["retval"], jnp.array([19.907951, 19.926113, 20.161331, 19.842987, 19.94448])
    )


def test_repeat_of_cond():
    repeated_model = cond_model(60.0).repeat(5)
    tr = repeated_model.simulate(key0)
    assert jnp.allclose(
        tr["retval"], jnp.array([60.057796, 31.760138, 30.233942, 31.401255, 60.03401])
    )


def test_vmap():
    @Gen
    def model(x, y):
        return x + Normal(y, 0.01) @ "a"

    tr = model(5.0, 1.0).simulate(key0)
    assert tr["retval"] == 5.9979415

    gf = model.vmap(in_axes=(0, None))(jnp.arange(5.0), 1.0)
    tr0 = gf.simulate(key0)

    assert jnp.allclose(
        tr0["retval"],
        jnp.array([0.99079514, 1.9926113, 3.016133, 3.9842987, 4.994448]),
    )
    tr1 = model.vmap(in_axes=(None, 0))(5.0, jnp.arange(0.1, 0.4, 0.1)).simulate(key0)
    assert jnp.allclose(
        tr1["retval"], jnp.array([5.1030984, 5.186357, 5.322861, 5.406714])
    )
    tr2 = jax.jit(
        model.vmap(in_axes=(0, 0))(
            jnp.arange(5.0), 0.1 * (1.0 + jnp.arange(5.0))
        ).simulate
    )(key0)
    assert jnp.allclose(
        tr2["retval"],
        jnp.array([0.09079514, 1.1926112, 2.316133, 3.3842988, 4.494448]),
    )
    # try the above without enumerating axis/arguments in in_axes
    tr3 = model.vmap()(jnp.arange(5.0), 0.1 * (1.0 + jnp.arange(5.0))).simulate(key0)
    assert jnp.allclose(tr3["retval"], tr2["retval"])


@pytest.mark.skip(reason="nested vmap not working yet")
def test_vmap_of_vmap():
    @Gen
    def model(x, y):
        return Normal(x, y) @ "n"

    tr = (
        model.vmap(in_axes=(0, None))
        # .vmap(in_axes=(None, 0))
        (jnp.arange(10.0, 15.0), jnp.arange(0.0, 1.6, 0.2)).simulate(key0)
    )
    assert jnp.allclose(
        tr["retval"],
        jnp.array(
            # this looked promising but it was cheating by using broadcast semantics on the inner loop
            [
                [10.0, 11.0, 12.0, 13.0, 14.0],
                [9.807053, 11.150885, 11.985554, 12.91451, 13.865575],
                [9.949703, 11.00537, 12.350649, 13.016957, 14.176276],
                [9.51422, 10.44197, 10.95027, 12.562811, 14.896659],
                [9.411664, 11.082625, 11.907326, 14.153446, 13.8836355],
            ]
        ),
    )


def test_assess():
    @Gen
    def p():
        x = Normal(0.0, 1.0) @ "x"
        y = Normal(0.0, 1.0) @ "y"
        return x, y

    @Gen
    def q():
        return p() @ "p"

    constraints = {"x": 2.0, "y": 2.1}
    w = p().assess(constraints)
    assert w == -6.0428767
    w = q().assess({"p": constraints})
    assert w == -6.0428767


def test_bernoulli():
    @Gen
    def p():
        b = Bernoulli(probs=0.01) @ "b"
        c = Bernoulli(logits=-1) @ "c"
        return b, c

    tr = p().simulate(key0)
    assert tr["retval"] == (0, 0)
    assert tr["subtraces"]["b"]["score"] == math.log(1 - 0.01)
    assert tr["subtraces"]["c"]["score"] == math.log(
        1 - math.exp(-1) / (1 + math.exp(-1))
    )

    with pytest.raises(ValueError):
        Bernoulli()

    with pytest.raises(ValueError):
        Bernoulli(logits=-1, probs=0.5)


def test_importance():
    @Gen
    def model():
        a = Normal(0.0, 1.0) @ "a"
        b = Normal(0.0, 0.1) @ "b"
        return a, b

    @Gen
    def outer():
        c = Normal(0.0, 1.0) @ "c"
        a, b = model() @ "d"
        return a + b + c

    model_imp = jax.jit(model().importance)
    outer_imp = jax.jit(outer().importance)

    tr1 = model_imp(key0, {"a": 1.0})
    assert tr1["w"] == -1.4189385
    tr2 = model_imp(key0, {"b": 1.0})
    assert tr2["w"] == -48.616352
    tr3 = model_imp(key0, {"a": 1.0, "b": 1.0})
    assert tr3["w"] == tr1["w"] + tr2["w"]

    tr4 = outer_imp(key0, {"c": 0.5, "d": {"b": 0.3}})

    assert tr4["w"] == -4.160292


def test_repeat_importance():
    @Gen
    def model(z):
        a = Normal(z, 0.1) @ "a"
        b = Normal(z, 1.0) @ "b"
        return a + b

    mr = model(1.0).repeat(4)
    mr_imp = jax.jit(mr.importance)
    values = jnp.arange(4) / 10.0
    tr = mr_imp(key0, {"a": values})
    assert jnp.allclose(tr["subtraces"]["a"]["retval"], values)
    assert tr["w"] == -141.46541
    assert tr["w"] == jnp.sum(tr["subtraces"]["a"]["w"])


def test_vmap_importance():
    @Gen
    def model(z):
        a = Normal(z, 0.1) @ "a"
        b = Normal(z, 1.0) @ "b"
        return a + b

    values = jnp.arange(4.0)
    mv = model.vmap()(values)
    mv_imp = jax.jit(mv.importance)
    observed_values = values + 0.1
    values = jnp.array(observed_values)
    tr = mv_imp(key0, {"a": values})
    assert tr["w"] == 3.534588
    assert jnp.allclose(tr["subtraces"]["a"]["retval"], observed_values)


# %%



import dataclasses
import jax.tree_util


class IterablePytree:
    def __getitem__(self, i):
        return jax.tree.map(lambda v: v[i], self)

    def __size__(self):
        return len(jax.tree.leaves(self)[0])

    def __iter__(self):
        return (self.__getitem__(i) for i in range(self.__size__()))


def pytree[R](cls: type) -> type:
    T = type(
        cls.__name__,
        (dataclasses.dataclass(cls), IterablePytree),
        {},
    )
    return jax.tree_util.register_dataclass(T)



from .minigenjax import (
    Gen,
    Normal,
    Uniform,
    Cond,
    Flip,
    to_constraint,
    Constraint,
    Categorical,
    Bernoulli,
    MvNormalDiag,
    Scan,
)
from .pytree import pytree as pytree

__all__ = [
    "Gen",
    "Normal",
    "Uniform",
    "Cond",
    "Flip",
    "to_constraint",
    "Constraint",
    "Categorical",
    "Bernoulli",
    "MvNormalDiag",
    "Scan",
    "pytree",
]



# %%
from typing import Any, Callable, Sequence
from jax._src.core import ClosedJaxpr as ClosedJaxpr
import tensorflow_probability.substrates.jax as tfp
import jax
import jax.tree
import jax.api_util
import jax.numpy as jnp
from jax.interpreters import batching, mlir

import jax.extend as jx
import jax.core
from jaxtyping import Array, ArrayLike, PRNGKeyArray


# %%
Address = tuple[str, ...]
Constraint = dict[str, "ArrayLike|Constraint"]
PHANTOM_KEY = jax.random.key(987654321)

WrappedFunWithAux = tuple[jx.linear_util.WrappedFun, Callable[[], Any]]

# Wrapper to assign a correct type.
flatten_fun_nokwargs: Callable[[jx.linear_util.WrappedFun, Any], WrappedFunWithAux] = (
    jax.api_util.flatten_fun_nokwargs  # pyright: ignore[reportAssignmentType]
)


class GenPrimitive(jx.core.Primitive):
    def __init__(self, name):
        super().__init__(name)
        self.def_abstract_eval(self.abstract)
        self.def_impl(self.concrete)
        batching.primitive_batchers[self] = self.batch

    def abstract(self, *args, **kwargs):
        raise NotImplementedError(f"abstract: {self}")

    def concrete(self, *args, **kwargs):
        raise NotImplementedError(f"concrete: {self}")

    def batch(self, vector_args, batch_axes, **kwargs):
        # TODO assert all axes equal
        result = jax.vmap(lambda *args: self.impl(*args, **kwargs), in_axes=batch_axes)(
            *vector_args
        )
        batched_axes = (
            (batch_axes[0],) * len(result) if self.multiple_results else batch_axes[0]
        )
        return result, batched_axes

    def simulate_p(
        self,
        key: PRNGKeyArray,
        arg_tuple: tuple,
        address: Address,
        constraint: Constraint,
    ) -> dict:
        raise NotImplementedError(f"simulate_p: {self}")

    def assess_p(
        self, arg_tuple: tuple, constraint: Constraint, address: tuple[str, ...]
    ) -> tuple[Array, Array]:
        raise NotImplementedError(f"assess_p: {self}")

    def inflate(self, v: Any, n: int):
        def inflate_one(v):
            return v.update(shape=(n,) + v.shape)

        if isinstance(v, tuple):
            return tuple(map(inflate_one, v))
        return inflate_one(v)


InAxesT = int | Sequence[Any] | None


class GFI[R](GenPrimitive):
    def __init__(self, name):
        super().__init__(name)

    def simulate(self, key: PRNGKeyArray) -> dict:
        return self.simulate_p(key, self.get_args(), (), {})

    def importance(self, key: PRNGKeyArray, constraint: Constraint) -> dict:
        return self.simulate_p(key, self.get_args(), (), constraint)

    def assess(self, constraint) -> Array:
        return self.assess_p(self.get_args(), constraint, ())[1]

    def __matmul__(self, address: str) -> R:
        raise NotImplementedError(f"{self} @ {address}")

    def get_args(self) -> tuple:
        raise NotImplementedError(f"get_args: {self}")

    def map[S](self, f: Callable[[R], S]) -> "MapGF[R,S]":
        return MapGF(self, f)

    def repeat(self, n: int) -> "RepeatGF[R]":
        return RepeatGF(self, n)

    def get_jaxpr(self) -> jx.core.ClosedJaxpr:
        raise NotImplementedError(f"get_jaxpr: {self}")

    def get_structure(self) -> jax.tree_util.PyTreeDef:
        raise NotImplementedError(f"get_structure: {self}")

    @staticmethod
    def make_jaxpr(f, arg_tuple):
        flat_args, in_tree = jax.tree.flatten(arg_tuple)
        flat_f, out_tree = flatten_fun_nokwargs(jx.linear_util.wrap_init(f), in_tree)
        # TODO: consider whether we need shape here
        jaxpr, shape = jax.make_jaxpr(flat_f.call_wrapped, return_shape=True)(
            *flat_args
        )
        structure = out_tree()
        return jaxpr, flat_args, structure


class Distribution(GenPrimitive):
    def __init__(self, name, tfd_ctor):
        super().__init__(name)
        self.tfd_ctor = tfd_ctor
        mlir.register_lowering(self, mlir.lower_fun(self.impl, False))

    def abstract(self, *args, **kwargs):
        return args[1]

    def concrete(self, *args, **kwargs):
        match kwargs.get("op", "Sample"):
            case "Sample":
                # we convert to float here because Bernoulli/Flip will
                # normally return an int, and that confuses XLA, since
                # our abstract implementation says the return types are
                # floats. TODO: consider allowing the marking of integer-
                # returning distributions as ints are sometimes nice to
                # work with.
                return jnp.asarray(
                    self.tfd_ctor(*args[1:]).sample(seed=args[0]), dtype=float
                )
            case "Score":
                return self.tfd_ctor(*args[1:]).log_prob(args[0])
            case _:
                raise NotImplementedError(f"{self.name}.{kwargs['op']}")

    def __call__(self, *args):
        this = self

        class Binder:
            def __matmul__(self, address: str):
                return this.bind(PHANTOM_KEY, *args, at=address)

            def __call__(self, key: PRNGKeyArray):
                return this.tfd_ctor(*args).sample(seed=key)

        return Binder()


BernoulliL = Distribution(
    "Bernoulli:L",
    lambda logits: tfp.distributions.Bernoulli(logits=logits),
)
BernoulliP = Distribution(
    "Bernoulli:P",
    lambda probs: tfp.distributions.Bernoulli(probs=probs),
)
Normal = Distribution("Normal", tfp.distributions.Normal)
MvNormalDiag = Distribution("MvNormalDiag", tfp.distributions.MultivariateNormalDiag)
Uniform = Distribution("Uniform", tfp.distributions.Uniform)
Flip = Distribution("Flip", lambda p: tfp.distributions.Bernoulli(probs=p))
CategoricalL = Distribution(
    "Categorical:L",
    lambda logits: tfp.distributions.Categorical(logits=logits),
)
CategoricalP = Distribution(
    "Categorical:P",
    lambda probs: tfp.distributions.Categorical(probs=probs),
)


def choose_scale(logits, probs, logit_dist, prob_dist):
    if (logits is None) == (probs is None):
        raise ValueError("Supply exactly one of logits=, probs=")
    return logit_dist(logits) if logits is not None else prob_dist(probs)


def Bernoulli(*, logits=None, probs=None):
    return choose_scale(logits, probs, BernoulliL, BernoulliP)


def Categorical(*, logits=None, probs=None):
    return choose_scale(logits, probs, CategoricalL, CategoricalP)


class KeySplitP(jx.core.Primitive):
    KEY_TYPE = jax.core.get_aval(PHANTOM_KEY)  # jax.core.ShapedArray((2,), jnp.uint32)

    def __init__(self):
        super().__init__("KeySplit")

        def impl(k, a=None):
            r = jax.random.split(k, 2)
            return r[0], r[1]

        self.def_impl(impl)
        self.multiple_results = True
        self.def_abstract_eval(
            lambda _, a=None: [KeySplitP.KEY_TYPE, KeySplitP.KEY_TYPE]
        )

        mlir.register_lowering(self, mlir.lower_fun(self.impl, self.multiple_results))

        batching.primitive_batchers[self] = self.batch

    def batch(self, vector_args, batch_axes, a=None):
        # key_pair_vector = jax.vmap(self.impl, in_axes=batch_axes)(*vector_args)
        v0, v1 = jax.vmap(self.impl, in_axes=batch_axes)(*vector_args)
        return [v0, v1], (batch_axes[0], batch_axes[0])


KeySplit = KeySplitP()
GenSymT = Callable[[jax.core.AbstractValue], jx.core.Var]


# %%
class Gen[R]:
    def __init__(self, f: Callable[..., R]):
        self.f = f

    def __call__(self, *args) -> "GF[R]":
        return GF(self.f, args)

    def vmap(self, in_axes: InAxesT = 0) -> Callable[..., "VmapGF[R]"]:
        return lambda *args: VmapGF(self, args, in_axes)


class GF[R](GFI[R]):
    def __init__(self, f: Callable[..., R], args: tuple):
        super().__init__(f"GF[{f.__name__}]")
        self.f = f

        self.args = args
        self.jaxpr, self.flat_args, self.structure = self.make_jaxpr(self.f, self.args)
        self.multiple_results = self.structure.num_leaves > 1

        a_vals = [ov.aval for ov in self.jaxpr.jaxpr.outvars]
        self.abstract_value = a_vals if self.multiple_results else a_vals[0]

    def abstract(self, *args, **_kwargs):
        return self.abstract_value

    def simulate_p(
        self,
        key: PRNGKeyArray,
        arg_tuple: tuple,
        address: Address,
        constraint: Constraint,
    ) -> dict:
        return Simulate(key, address, constraint).run(
            self.jaxpr, arg_tuple, self.structure
        )

    def assess_p(
        self, arg_tuple: tuple, constraint: Constraint, address
    ) -> tuple[Array, Array]:
        a = Assess(address, constraint)
        value = a.run(self.jaxpr, arg_tuple)
        return value, a.score

    def __matmul__(self, address: str):
        return self.bind(*self.flat_args, at=address)

    def get_args(self) -> tuple:
        return self.args

    def get_jaxpr(self) -> jx.core.ClosedJaxpr:
        return self.jaxpr

    def get_structure(self) -> jax.tree_util.PyTreeDef:
        return self.structure


class Transformation[R]:
    def __init__(self, key: PRNGKeyArray, address: Address, constraint: Constraint):
        self.key = key
        self.address = address
        self.constraint = constraint

    def handle_eqn(self, eqn, params, bind_params):
        return eqn.primitive.bind(*params, **bind_params)

    def get_sub_key(self):
        if self.key_consumer_count > 1:
            self.key, sub_key = KeySplit.bind(self.key)
        elif self.key_consumer_count == 1:
            sub_key = self.key
        else:
            raise Exception("more sub_key requests than expected")
        self.key_consumer_count -= 1
        return sub_key

    def run(
        self,
        closed_jaxpr: jx.core.ClosedJaxpr,
        arg_tuple,
        structure: jax.tree_util.PyTreeDef | None = None,
    ):
        jaxpr = closed_jaxpr.jaxpr
        flat_args, in_tree = jax.tree.flatten(arg_tuple)
        env: dict[jx.core.Var, Any] = {}

        def read(v: jax.core.Atom) -> Any:
            return v.val if isinstance(v, jx.core.Literal) else env[v]

        def write(v: jx.core.Var, val: Any) -> None:
            # if config.enable_checks.value and not config.dynamic_shapes.value:
            #   assert typecheck(v.aval, val), (v.aval, val)
            env[v] = val

        jax.util.safe_map(write, jaxpr.constvars, closed_jaxpr.consts)
        jax.util.safe_map(write, jaxpr.invars, flat_args)

        # count the number of PRNG keys that will be consumed during the
        # evaluation of this JAXPR alone. We assume that the key that we
        # were provided with is good for one random number generation. If
        # there's only one key consumer in this JAXPR, then there's no need
        # to split it.
        self.key_consumer_count = sum(
            isinstance(eqn.primitive, GenPrimitive)
            or eqn.primitive is jax.lax.cond_p
            or eqn.primitive is jax.lax.scan_p
            for eqn in jaxpr.eqns
        )

        for eqn in jaxpr.eqns:
            sub_fns, bind_params = eqn.primitive.get_bind_params(eqn.params)
            if sub_fns:
                raise NotImplementedError("nonempty sub_fns")
            # name_stack = source_info_util.current_name_stack() + eqn.source_info.name_stack
            # traceback = eqn.source_info.traceback if propagate_source_info else None
            # with source_info_util.user_context(
            #    traceback, name_stack=name_stack), eqn.ctx.manager:
            params = tuple(jax.util.safe_map(read, eqn.invars))
            ans = self.handle_eqn(eqn, params, bind_params)
            if eqn.primitive.multiple_results:
                jax.util.safe_map(write, eqn.outvars, ans)
            else:
                write(eqn.outvars[0], ans)
            # clean_up_dead_vars(eqn, env, lu)
        retval = jax.util.safe_map(read, jaxpr.outvars)
        if structure is not None:
            retval = jax.tree.unflatten(structure, retval)
        else:
            retval = retval if len(jaxpr.outvars) > 1 else retval[0]
        return self.construct_retval(retval)

    def address_from_branch(self, b: jx.core.ClosedJaxpr):
        """Look at the given JAXPR and find out if it is a single-instruction
        call to a GF traced to an address. If so, return that address. This is
        used to detect when certain JAX primitives (e.g., `scan_p`, `cond_p`)
        have been applied directly to traced generative functions, in which case
        the current transformation should be propagated to the jaxpr within."""
        if len(b.jaxpr.eqns) == 1 and isinstance(b.jaxpr.eqns[0].primitive, GF):
            return b.jaxpr.eqns[0].params.get("at")

    def apply_constraint(self, addr: str) -> ArrayLike | None:
        if isinstance(self.constraint, dict):
            v = self.constraint.get(addr)
            if isinstance(v, dict):
                raise Exception(
                    "broken constraint: {addr} in {self.address} is not a leaf"
                )
            return v
        else:
            return self.constraint

    def get_sub_constraint(self, a: str) -> Constraint:
        d = self.constraint.get(a, {})
        if not isinstance(d, dict):
            return {}
        return d

    def construct_retval(self, retval) -> R:
        return retval


class Assess(Transformation[Array]):
    def __init__(self, address: Address, constraint: Constraint):
        super().__init__(PHANTOM_KEY, address, constraint)
        self.score = jnp.array(0.0)

    def handle_eqn(self, eqn: jx.core.JaxprEqn, params, bind_params):
        if isinstance(eqn.primitive, Distribution):
            addr = self.address + (bind_params["at"],)
            v = self.apply_constraint(bind_params["at"])
            self.score += eqn.primitive.bind(v, *params[1:], op="Score")
            return v

        if isinstance(eqn.primitive, GenPrimitive):
            a = bind_params["at"]
            addr = self.address + (a,)
            sub_constraint = self.constraint.get(a, {})
            assert isinstance(sub_constraint, dict)
            ans, score = eqn.primitive.assess_p(params, sub_constraint, addr)
            self.score += score
            return ans

        return super().handle_eqn(eqn, params, bind_params)


class Simulate(Transformation[dict]):
    def __init__(
        self,
        key: PRNGKeyArray,
        address: Address,
        constraint: Constraint,
    ):
        super().__init__(key, address, constraint)
        self.trace = {}
        self.w = jnp.array(0.0)

    def transform_inner(
        self, jaxpr, in_avals, addr: str, constraint: Constraint | None = None
    ):
        """Apply simulate to jaxpr and return the transformed jaxpr together
        with its return shape."""

        if constraint:
            reduced_constraint = jax.tree.map(lambda v: v[0], constraint)
            return jax.make_jaxpr(
                lambda key, co, in_avals: Simulate(key, self.address + (addr,), co).run(
                    jaxpr, in_avals
                ),
                return_shape=True,
            )(PHANTOM_KEY, reduced_constraint, in_avals)
        else:
            return jax.make_jaxpr(
                lambda key, in_avals: Simulate(key, self.address + (addr,), {}).run(
                    jaxpr, in_avals
                ),
                return_shape=True,
            )(PHANTOM_KEY, in_avals)

    def record(self, subtrace, at):
        if (
            (inner_trace := subtrace.get("subtraces"))
            and len(keys := inner_trace.keys()) == 1
            and (key := next(iter(keys))).startswith("__")
        ):
            # absorb interstitial trace points like __repeat, __vmap that may
            # occur when combinators are stacked. Preseve the outermost retval,
            # though, as it will have the correct structure (?) TODO: is this true?
            subtrace["subtraces"] = inner_trace[key]["subtraces"]
        if at:
            self.trace[at] = subtrace
        self.w += jnp.sum(subtrace.get("w", 0.0))
        return subtrace["retval"]

    def handle_eqn(self, eqn: jx.core.JaxprEqn, params, bind_params):
        if isinstance(eqn.primitive, RepeatGF):
            at = bind_params["at"]
            sub_constraint = self.get_sub_constraint(at)
            transformed, shape = self.transform_inner(
                bind_params["inner"], params, at, sub_constraint
            )
            new_params = bind_params | {"inner": transformed}
            has_constraint = bool(sub_constraint)
            if has_constraint:
                flat_constraint, _ = jax.tree.flatten(sub_constraint)
                ans = eqn.primitive.Simulate(eqn.primitive, shape, has_constraint).bind(
                    self.get_sub_key(), *flat_constraint, *params, **new_params
                )
            else:
                ans = eqn.primitive.Simulate(eqn.primitive, shape, has_constraint).bind(
                    self.get_sub_key(), *params, **new_params
                )

            u = jax.tree.unflatten(jax.tree.structure(shape), ans)
            return self.record(u, at)

        if isinstance(eqn.primitive, VmapGF):
            at = bind_params["at"]
            sub_constraint = self.get_sub_constraint(at)
            transformed, shape = self.transform_inner(
                bind_params["inner"],
                eqn.primitive.reduced_avals(params),
                at,
                sub_constraint,
            )
            new_params = bind_params | {"inner": transformed}
            has_constraint = bool(sub_constraint)
            flat_params, _ = jax.tree.flatten(params)
            if has_constraint:
                flat_constraint, _ = jax.tree.flatten(sub_constraint)
                ans = eqn.primitive.Simulate(eqn.primitive, shape, has_constraint).bind(
                    self.get_sub_key(), *flat_constraint, *flat_params, **new_params
                )
            else:
                ans = eqn.primitive.Simulate(eqn.primitive, shape, has_constraint).bind(
                    self.get_sub_key(), *flat_params, **new_params
                )
            u = jax.tree.unflatten(jax.tree.structure(shape), ans)
            if z := u["subtraces"].get(RepeatGF.SUB_TRACE):
                u = z
            return self.record(u, at)

        # we regard the presence of constraints as enabling importance mode,
        # so you can't do an importance over empty constraints to get a zero
        # score; instead, that will be interpreted as a regular simulate.
        if isinstance(eqn.primitive, Distribution):
            at = bind_params["at"]
            addr = self.address + (at,)
            if (retval := self.apply_constraint(at)) is not None:
                score = eqn.primitive.bind(retval, *params[1:], op="Score")
                ans = {"w": score, "retval": retval}
            else:
                retval = eqn.primitive.bind(
                    self.get_sub_key(), *params[1:], op="Sample"
                )
                score = eqn.primitive.bind(retval, *params[1:], op="Score")
                ans = {"retval": retval, "score": score}
            return self.record(ans, at)

        if isinstance(eqn.primitive, GenPrimitive):
            at = bind_params["at"]
            addr = self.address + (at,)
            ans = eqn.primitive.simulate_p(
                self.get_sub_key(), params, addr, self.get_sub_constraint(at)
            )
            return self.record(ans, at)

        if eqn.primitive is jax.lax.cond_p:
            branches = bind_params["branches"]

            branch_addresses = tuple(map(self.address_from_branch, branches))
            if branch_addresses[0] and all(
                b == branch_addresses[0] for b in branch_addresses[1:]
            ):
                sub_address = branch_addresses[0]
            else:
                sub_address = None

            # TODO: is it OK to pass the same sub_key to both sides?
            # NB! branches[0] is the false branch, [1] is the true branch,
            sub_key = self.get_sub_key()
            ans = jax.lax.cond(
                params[0],
                lambda: Simulate(sub_key, self.address, self.constraint).run(
                    branches[1], params[1:]
                ),
                lambda: Simulate(sub_key, self.address, self.constraint).run(
                    branches[0], params[1:]
                ),
            )
            if sub_address:
                self.trace[sub_address] = ans["subtraces"][sub_address]

            self.w += jnp.sum(ans.get("w", 0))
            # The reasons why this result has to be a sequence is obscure to me,
            # but cond_p as a primitive requires "multiple results."
            return (ans["retval"],)

        if eqn.primitive is jax.lax.scan_p:
            inner = bind_params["jaxpr"]

            if at := self.address_from_branch(inner):
                address = self.address + (at,)
                sub_address = at
            else:
                address = self.address
                sub_address = None

            def step(carry_key, s):
                carry, key = carry_key
                key, k1 = KeySplit.bind(key, a="scan_step")
                v = Simulate(k1, address, self.constraint).run(inner, (carry, s))
                return ((v["retval"][0], key), v)

            ans = jax.lax.scan(step, (params[0], self.get_sub_key()), params[1])
            if sub_address:
                self.trace[sub_address] = ans[1]["subtraces"][sub_address]

            self.w += jnp.sum(ans[1].get("w", 0))
            # we extended the carry with the key; now drop it
            return (ans[0][0], ans[1]["retval"][1])

        return super().handle_eqn(eqn, params, bind_params)

    def construct_retval(self, retval):
        r = {"retval": retval, "subtraces": self.trace}
        if self.constraint:
            r["w"] = self.w
        return r


# %%
def Cond(tf, ff):
    """Cond combinator. Turns (tf, ff) into a function of a boolean
    argument which will switch between the true and false branches."""

    def ctor(pred):
        pred_as_int = jnp.int32(pred)

        class Binder:
            def __matmul__(self, address: str):
                return jax.lax.switch(
                    pred_as_int, [lambda: ff @ address, lambda: tf @ address]
                )

            def simulate(self, key: PRNGKeyArray):
                return GF(lambda: self @ "__cond", ()).simulate(key)

        return Binder()

    return ctor


def Scan(gf: Gen):
    """Scan combinator. Turns a GF of two parameters `(state, update)`
    returning a pair of updated state and step data to record into
    a generative function of an initial state and an array of updates."""

    def ctor(init, steps):
        class Binder:
            def __matmul__(self, address: str):
                def inner(carry, step):
                    c, s = gf(carry, step) @ address
                    return c, s

                return jax.lax.scan(inner, init, steps)

        return Binder()

    return ctor


class RepeatGF[R](GFI[R]):
    SUB_TRACE = "__repeat"

    def __init__(self, gfi: GFI[R], n: int):
        super().__init__(f"Repeat[{gfi.name}, {n}]")
        self.gfi = gfi
        self.n = n
        self.multiple_results = self.gfi.multiple_results
        # TODO: try to reuse self.__matmul__ here, if possible
        self.jaxpr, self.flat_args, self.structure = self.make_jaxpr(
            lambda *args: self.bind(
                *args, at=RepeatGF.SUB_TRACE, n=self.n, inner=self.gfi.get_jaxpr()
            ),
            self.get_args(),
        )

    def abstract(self, *args, **kwargs):
        return self.inflate(self.gfi.abstract(*args, **kwargs), self.n)

    def simulate_p(
        self,
        key: PRNGKeyArray,
        arg_tuple: tuple,
        address: Address,
        constraint: Constraint,
    ) -> dict:
        tr = Simulate(key, address, {RepeatGF.SUB_TRACE: constraint}).run(
            self.jaxpr, arg_tuple, self.structure
        )["subtraces"][RepeatGF.SUB_TRACE]
        if "w" in tr:
            tr["w"] = jnp.sum(tr["w"])
        return tr

    def __matmul__(self, address: str) -> R:
        return self.bind(
            *self.gfi.get_args(), at=address, n=self.n, inner=self.gfi.get_jaxpr()
        )

    def get_jaxpr(self) -> jx.core.ClosedJaxpr:
        return self.jaxpr

    def get_structure(self):
        return self.structure

    def get_args(self) -> tuple:
        return self.gfi.get_args()

    class Simulate[S](GFI[S]):
        def __init__(self, r: "RepeatGF[S]", shape: Any, has_constraint: bool):
            super().__init__("Repeat.Simulate")
            self.r = r
            self.shape = shape
            self.has_constraint = has_constraint
            self.multiple_results = True
            mlir.register_lowering(
                self, mlir.lower_fun(self.impl, self.multiple_results)
            )

        def abstract(self, *args, **kwargs):
            return [
                self.inflate(jax.core.get_aval(s), self.r.n)
                for s in jax.tree.flatten(self.shape)[0]
            ]

        def concrete(self, *args, **kwargs):
            # this is called after the simulate transformation so the key is the first argument
            j: jx.core.ClosedJaxpr = kwargs["inner"]
            if self.has_constraint:
                # TODO: there could be several constraint arguments; have a param count how many
                return jax.vmap(
                    lambda k, constraint, arg_tuple: jax.core.eval_jaxpr(
                        j.jaxpr, j.consts, k, constraint, *arg_tuple
                    ),
                    in_axes=(0, 0, None),
                )(jax.random.split(args[0], kwargs["n"]), args[1], args[2:])
            return jax.vmap(
                lambda k: jax.core.eval_jaxpr(j.jaxpr, j.consts, k, *args[1:]),
                in_axes=(0,),
            )(jax.random.split(args[0], kwargs["n"]))


class VmapGF[R](GFI[R]):
    SUB_TRACE = "__vmap"

    def __init__(self, g: Gen, arg_tuple: tuple, in_axes: InAxesT):
        super().__init__(f"Vmap[{g.f.__name__}]")
        if in_axes is None or in_axes == ():
            raise NotImplementedError(
                "must specify at least one argument/axis for Vmap"
            )
        # TODO: consider if we want to make this
        self.arg_tuple = arg_tuple
        self.flat_args, self.in_tree = jax.tree.flatten(self.arg_tuple)
        self.flat_args = tuple(self.flat_args)
        self.in_axes = in_axes
        # find one pair of (parameter number, axis) to use to determine size of vmap
        if isinstance(self.in_axes, tuple):
            self.p_index, self.an_axis = next(
                filter(lambda b: b[1] is not None, enumerate(self.in_axes))
            )
        else:
            self.p_index, self.an_axis = 0, self.in_axes
        # Compute the "scalar" jaxpr by feeding the un-v-mapped arguments to make_jaxpr
        # TODO: does `self` need to remember these?
        self.inner_jaxpr, self.inner_shape = jax.make_jaxpr(g.f, return_shape=True)(
            *jax.tree.unflatten(self.in_tree, self.reduced_avals(self.arg_tuple))
        )
        n = self.arg_tuple[self.p_index].shape[self.an_axis]
        # the shape of the v-mapped function will increase every axis of the result
        self.shape = jax.ShapeDtypeStruct(
            (n,) + self.inner_shape.shape, self.inner_shape.dtype
        )
        self.jaxpr, self.shape = jax.make_jaxpr(
            lambda *args: self.bind(
                *args, in_axes=self.in_axes, at=VmapGF.SUB_TRACE, inner=self.inner_jaxpr
            ),
            return_shape=True,
        )(*self.flat_args)

    def reduced_avals(self, arg_tuple):
        # if in_axes is not an tuple, lift it to the same shape as arg tuple
        if isinstance(self.in_axes, int):
            ia = jax.tree.map(lambda _: self.in_axes, arg_tuple)
        else:
            ia = self.in_axes

        # Now produce an abstract arg tuple in which the shape of the
        # arrays is contracted in those position where vmap would expect
        # to find a mapping axis
        def deflate(array, axis):
            if axis is None:
                return array
            aval = jax.core.get_aval(array)
            # delete the indicated axes from teh shape tuple
            assert isinstance(aval, jax.core.ShapedArray)
            return aval.update(shape=aval.shape[:axis] + aval.shape[axis + 1 :])

        return jax.tree.map(deflate, self.flat_args, ia)

    def abstract(self, *args, **kwargs):
        return self.shape

    def simulate_p(
        self,
        key: PRNGKeyArray,
        arg_tuple: tuple,
        address: Address,
        constraint: Constraint,
    ) -> dict:
        tr = Simulate(key, address, {VmapGF.SUB_TRACE: constraint}).run(
            self.jaxpr, arg_tuple
        )["subtraces"][VmapGF.SUB_TRACE]
        if "w" in tr:
            tr["w"] = jnp.sum(tr["w"])
        return tr

    def __matmul__(self, address: str) -> R:
        return self.bind(
            *self.flat_args, at=address, in_axes=self.in_axes, inner=self.inner_jaxpr
        )

    def get_args(self) -> tuple:
        return self.arg_tuple

    def get_jaxpr(self) -> jx.core.ClosedJaxpr:
        return self.jaxpr

    class Simulate[S](GFI[S]):
        def __init__(
            self, r: "VmapGF[S]", shape: Any, has_constraint: bool
        ):  # TODO: PyTreeDef?
            super().__init__("Vmap.Simulate")
            self.r = r
            self.n = self.r.arg_tuple[self.r.p_index].shape[self.r.an_axis]
            self.shape = shape
            self.has_constraint = has_constraint
            self.multiple_results = True
            mlir.register_lowering(
                self, mlir.lower_fun(self.impl, self.multiple_results)
            )

        def abstract(self, *args, **kwargs):
            return [
                self.inflate(jax.core.get_aval(s), self.n)
                for s in jax.tree.flatten(self.shape)[0]
            ]

        def concrete(self, *args, **kwargs):
            # this is called after the simulate transformation so the key is the first argument
            j: jx.core.ClosedJaxpr = kwargs["inner"]
            if self.has_constraint:
                # TODO: there could be several constraint arguments; have a param count how many
                return jax.vmap(
                    lambda k, constraint, arg_tuple: jax.core.eval_jaxpr(
                        j.jaxpr, j.consts, k, constraint, *arg_tuple
                    ),
                    in_axes=(0, 0, self.r.in_axes),
                )(jax.random.split(args[0], self.n), args[1], args[2:])

            return jax.vmap(
                lambda k, arg_tuple: jax.core.eval_jaxpr(
                    j.jaxpr, j.consts, k, *arg_tuple
                ),
                in_axes=(0, self.r.in_axes),
            )(jax.random.split(args[0], self.n), args[1:])


class MapGF[R, S](GFI[S]):
    def __init__(self, gfi: GFI[R], f: Callable[[R], S]):
        super().__init__(f"Map[{gfi.name}, {f.__name__}]")
        self.gfi = gfi
        self.f = f

        inner = self.gfi.get_jaxpr()
        self.jaxpr, self.flat_args, self.structure = self.make_jaxpr(
            lambda *args: self.f(
                jax.tree.unflatten(
                    self.gfi.get_structure(),
                    jax.core.eval_jaxpr(inner.jaxpr, inner.consts, *args),
                )
            ),
            self.gfi.get_args(),
        )

    def abstract(self, *args, **kwargs):
        # this can't be right: what about the effect of f? Why isn't it
        # sufficient to apply f to a tracer to compose the abstraction?
        return self.gfi.abstract(*args, **kwargs)

    def simulate_p(
        self,
        key: PRNGKeyArray,
        arg_tuple: tuple,
        address: Address,
        constraint: Constraint,
    ) -> dict:
        print(f"map simulating with structure {self.structure}")
        return Simulate(key, address, constraint).run(
            self.jaxpr, arg_tuple, self.structure
        )
        # v = self.gfi.simulate_p(key, arg_tuple, address, constraint)
        # v["retval"] = self.f(v["retval"])
        # return v

    def __matmul__(self, address: str) -> S:
        # TODO (Q): can we declare multiple returns and drop the brackets?
        return jax.tree.unflatten(
            self.structure, [self.bind(*self.flat_args, at=address)]
        )

    def get_args(self) -> tuple:
        return self.gfi.get_args()

    def get_jaxpr(self) -> jx.core.ClosedJaxpr:
        return self.jaxpr


def to_constraint(trace: dict) -> Constraint:
    if "subtraces" in trace:
        return {k: to_constraint(v) for k, v in trace["subtraces"].items()}
    return trace["retval"]
